class Abbreviations:
    def abbr(self, abbr="", string="", **kwargs):
        """Defines an abbreviation.

        APDL Command: ``*ABBR``

        Parameters
        ----------
        abbr
            The abbreviation (up to 8 alphanumeric characters) used to
            represent the string String.  If Abbr is the same as an existing
            ANSYS command, the abbreviation overrides.  Avoid using an Abbr
            which is the same as an ANSYS command.

        string
            String of characters (60 maximum) represented by Abbr.
            Cannot include a $ or any of the commands ``C***``, /COM,
            /GOPR, /NOPR, /QUIT, /UI, or ``*END``.

            Use the ``*IF`` groups may not be abbreviated. If String
            is blank, the abbreviation is deleted. To abbreviate
            multiple commands, create an "unknown command" macro or
            define String to execute a macro file [``*USE``]
            containing the desired commands.

        Notes
        -----
        Once the abbreviation Abbr is defined, you can issue it at the
        beginning of a command line and follow it with a blank (or with a comma
        and appended data), and the program will substitute the string  String
        for Abbr as the line is executed. Up to 100 abbreviations may exist at
        any time and are available throughout the program. Abbreviations may be
        redefined or deleted at any time.

        Use ``*STATUS`` to display the current list of abbreviations. For
        abbreviations repeated with ``*REPEAT``, substitution occurs before the
        repeat increments are applied. There are a number of abbreviations that
        are predefined by the program (these can be deleted by using the blank
        String option described above). Note that String will be written to the
        File.LOG.

        This command is valid in any processor.
        """
        command = f"*ABBR,{abbr},{string}"
        return self.run(command, **kwargs)

    def abbres(self, lab="", fname="", ext="", **kwargs):
        """Reads abbreviations from a coded file.

        APDL Command: ABBRES

        Parameters
        ----------
        lab
            Label that specifies the read operation:

            Replace current abbreviation set with these abbreviations (default). - Extend current abbreviation set with these abbreviations, replacing any of the
                              same name that already exist.

        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        The abbreviation file may have been written with the ABBSAV command. Do
        not issue ABBRES,NEW while inside an executing abbreviation. Doing so
        will cause all data for the executing abbreviation to be deleted.

        This command is valid in any processor.
        """
        command = f"ABBRES,{lab},{fname},{ext}"
        return self.run(command, **kwargs)

    def abbsav(self, lab="", fname="", ext="", **kwargs):
        """Writes the current abbreviation set to a coded file.

        APDL Command: ABBSAV

        Parameters
        ----------
        lab
            Label that specifies the write operation.

            ALL - Write all abbreviations (default).

        fname
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).
            An unspecified directory path defaults to the working
            directory; in this case, you can use all 248 characters
            for the file name.

            The file name defaults to Jobname.

        ext
            Filename extension (eight-character maximum).  The
            extension defaults to ABBR if Fname is blank.

        Notes
        -----
        Existing abbreviations on this file, if any, will be overwritten.  The
        abbreviation file may be read with the ABBRES command.

        This command is valid in any processor.
        """
        command = f"ABBSAV,{lab},{fname},{ext}"
        return self.run(command, **kwargs)

    def ucmd(self, cmd="", srnum="", **kwargs):
        """Assigns a user-defined command name.

        APDL Command: /UCMD

        Parameters
        ----------
        cmd
            User-defined command name.  Only the first four characters are
            significant.  Must not conflict with any ANSYS command name or any
            user "unknown command" macro name.

        srnum
            User subroutine number (1 to 10) programmed for this command.  For
            example, the command /UCMD,MYCMD,3 will execute subroutine USER03
            whenever the command MYCMD is entered.  Use a blank command name to
            disassociate SRNUM from its command.  For example, /UCMD,,3 removes
            MYCMD as a command.

        Notes
        -----
        Assigns a user-defined command name to a user-programmable (system-
        dependent) subroutine.  This feature allows user-defined commands to be
        programmed into the ANSYS program.  Once programmed, this command can
        be input to the program like other commands, and can also be included
        in the ANSYS start-up file.  See ``*ULIB`` for another way of defining user
        commands.

        Up to 10 subroutines are available for user-defined commands (USER01 to
        USER10).  Users must have system permission, system access, and
        knowledge to write, compile, and link the appropriate subprocessors
        into the ANSYS program at the site where it is to be run.  All routines
        should be written in FORTRAN. For more information on FORTRAN compilers
        please refer to either the ANSYS, Inc. Windows Installation Guide or
        the ANSYS, Inc. Linux Installation Guide for details specific to your
        platform or operating system. The USER01 routine is commented and
        should be listed from the distribution media (system dependent) for
        more details.  Issue /UCMD,STAT to list all user-defined command names.
        Since a user-programmed command is a nonstandard use of the program,
        the verification of any ANSYS run incorporating these commands is
        entirely up to the user.  In any contact with ANSYS customer support
        regarding the performance of a custom version of the ANSYS program, you
        should explicitly state that a user programmable feature has been used.
        See the Advanced Analysis Guide for a general description of user-
        programmable features and Guide to User-Programmable Features for a
        detailed description of these features.

        This command is valid only at the Begin Level.
        """
        command = f"/UCMD,{cmd},{srnum}"
        return self.run(command, **kwargs)



class LoadStepOperations:
    def lsclear(self, lab="", **kwargs):
        """Clears loads and load step options from the database.

        APDL Command: LSCLEAR

        Parameters
        ----------
        lab
            Label identifying the data to be cleared:

            SOLID - Delete only solid model loads.

            FE - Delete only finite element loads.

            INER - Delete only inertia loads (ACEL, etc.).

            LFACT - Initialize only load factors (on DCUM, FCUM, SFCUM, etc.).

            LSOPT - Initialize only load step options.

            ALL - Delete all loads and initialize all load step options and load factors.

        Notes
        -----
        Loads are deleted, and load step options are initialized to their
        default values.

        This command is also valid in PREP7.
        """
        command = f"LSCLEAR,{lab}"
        return self.run(command, **kwargs)

    def lsdele(self, lsmin="", lsmax="", lsinc="", **kwargs):
        """Deletes load step files.

        APDL Command: LSDELE

        Parameters
        ----------
        lsmin, lsmax, lsinc
            Range of load step files to be deleted, from LSMIN to LSMAX in
            steps of LSINC.  LSMAX defaults to LSMIN, and LSINC defaults to 1.
            If LSMIN = ALL, all load step files are deleted (and LSMAX and
            LSINC are ignored).  The load step files are assumed to be named
            Jobname.Sn, where n is a number assigned by the LSWRITE command (01
            --09,10,11, etc.).  On systems with a 3-character limit on the
            extension, the "S" is dropped for numbers > 99.

        Notes
        -----
        Deletes load step files in the current directory (written by the
        LSWRITE command).

        This command is also valid in PREP7.
        """
        command = f"LSDELE,{lsmin},{lsmax},{lsinc}"
        return self.run(command, **kwargs)

    def lsread(self, lsnum="", **kwargs):
        """Reads load and load step option data into the database.

        APDL Command: LSREAD

        Parameters
        ----------
        lsnum
            Identification number of the load step file to be read.  Defaults
            to 1 + highest number read in the current session.  Issue
            LSREAD,STAT to list the current value of LSNUM.  Issue LSREAD,INIT
            to reset LSNUM to 1.  The load step files are assumed to be named
            Jobname.Sn, where n is a number assigned by the LSWRITE command (01
            --09,10,11, etc.).  On systems with a 3-character limit on the
            extension, the "S" is dropped for LSNUM > 99.

        Notes
        -----
        Reads load and load step option data from the load step file into the
        database.  LSREAD will not clear the database of all current loads.
        However, if a load is respecified with LSREAD, then it will overwrite
        the existing load. See the LSWRITE command to write load step files,
        and the LSDELE command to delete load step files.  LSREAD removes any
        existing SFGRAD specification.

        This command is also valid in PREP7.
        """
        command = f"LSREAD,{lsnum}"
        return self.run(command, **kwargs)

    def lssolve(self, lsmin="", lsmax="", lsinc="", **kwargs):
        """Reads and solves multiple load steps.

        APDL Command: LSSOLVE

        Parameters
        ----------
        lsmin, lsmax, lsinc
            Range of load step files to be read and solved, from
            ``lsmin`` to ``lsmax`` in steps of ``lsinc``.  ``lsmax``
            defaults to ``lsmin``, and ``lsinc`` defaults to 1. If
            ``lsmin`` is blank, a brief command description is
            displayed.  The load step files are assumed to be named
            Jobname.Sn, where n is a number assigned by the
            ``lswrite`` command (01--09, 10, 11, etc.).  On systems
            with a 3-character limit on the extension, the "S" is
            dropped for numbers > 99.

        Notes
        -----
        ``lssolve`` invokes an ANSYS macro to read and solve multiple
        load steps.  The macro loops through a series of load step
        files written by the LSWRITE command.  The macro file called
        by ``lssolve`` is called LSSOLVE.MAC.

        ``lssolve`` cannot be used with the birth-death option.

        ``lssolve`` is not supported for cyclic symmetry analyses.

        ``lssolve`` does not support restarts.

        Examples
        --------
        Write the load and load step option data to a file and solve
        it.  In this case, write the second load step.

        >>> mapdl.lswrite(2)
        >>> mapdl.lssolve(1, 2)

        """
        with self.non_interactive:
            self.run(f"LSSOLVE,{lsmin},{lsmax},{lsinc}", **kwargs)
        return self.last_response

    def lswrite(self, lsnum="", **kwargs):
        """Writes load and load step option data to a file.

        APDL Command: LSWRITE

        Parameters
        ----------
        lsnum
            Number to be assigned to the load step file name for identification
            purposes.  Defaults to 1 + highest LSNUM used in the current
            session.  Issue LSWRITE,STAT to list the current value of LSNUM.
            Issue LSWRITE,INIT to reset to 1.  The load step file will be named
            Jobname.Sn, where n is the specified LSNUM value (preceded by "0"
            for values 1-9).  On systems with a 3-character limit on the file
            name extension, the "S" is dropped for LSNUM > 99.

        Notes
        -----
        Writes all load and load step option data for the selected model to a
        load step file for later use.  LSWRITE does not capture changes made to
        real constants (R), material properties (MP), couplings (CP), or
        constraint equations (CE).

        Solid model loads will not be saved if the model is not meshed. Solid
        model loads, if any, are transferred to the finite element model. Issue
        LSCLEAR,FE to delete finite element loads.

        One file is written for each load step. Use the LSREAD command to read
        a single load step file, and the LSDELE command to delete load step
        files.  Use the LSSOLVE command to read and solve the load steps
        sequentially.

        Solution control commands are typically not written to the file unless
        you specifically change a default solution setting.

        LSWRITE does not support the following commands: DJ, FJ, GSBDATA,
        GSGDATA, ESTIF, EKILL, EALIVE, MPCHG, and OUTRES. These commands will
        not be written to the load step file.

        LSWRITE cannot be used with the birth-death option.

        This command is also valid in PREP7.
        """
        command = f"LSWRITE,{lsnum}"
        return self.run(command, **kwargs)



class BinDump:
    def dump(self, nstrt="", nstop="", **kwargs):
        """Dumps the contents of a binary file.

        APDL Command: DUMP

        Parameters
        ----------
        nstrt, nstop
            Dump file from record NSTRT (defaults to 1) to NSTOP (defaults to
            NSTRT).  If NSTRT = HEAD, dump only record 1 of the file (NSTOP and
            the format specification are ignored).  If NSTRT = ALL, dump the
            entire file.

        Notes
        -----
        Dumps the file named on the AUX2 FILEAUX2 command according the format
        specified on the FORM command.
        """
        command = f"DUMP,{nstrt},{nstop}"
        return self.run(command, **kwargs)

    def fileaux2(self, fname="", ident="", **kwargs):
        """Specifies the binary file to be dumped.

        APDL Command: FILEAUX2

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).
            An unspecified directory path defaults to the working
            directory; in this case, you can use all 248 characters
            for the file name.

        ident
            ANSYS filename identifier.  See the Basic Analysis Guide
            for file descriptions and identifiers.  If not an ANSYS
            identifier, Ident will be used as the filename extension.

        Notes
        -----
        Specifies the binary file to be dumped with the DUMP command.
        """
        command = f"FILEAUX2,{fname},{ident}"
        return self.run(command, **kwargs)

    def form(self, lab="", **kwargs):
        """Specifies the format of the file dump.

        APDL Command: FORM

        Parameters
        ----------
        lab
            Format:

            RECO - Basic record description only (minimum output) (default).

            TEN - Same as RECO plus the first ten words of each record.

            LONG - Same as RECO plus all words of each record.

        Notes
        -----
        Specifies the format of the file dump (from the DUMP command).
        """
        command = f"FORM,{lab}"
        return self.run(command, **kwargs)

    def ptr(self, loc="", base="", **kwargs):
        """Dumps the record of a binary file.

        APDL Command: PTR

        Parameters
        ----------
        loc, base
            Dump the file record starting at pointer LOC. BASE is the base
            pointer, and would be used if LOC is a relative pointer.

        Notes
        -----
        Dumps the record of the file named on the AUX2 FILEAUX2 command
        according the format specified on the FORM command.
        """
        command = f"PTR,{loc},{base}"
        return self.run(command, **kwargs)



import os
import tempfile
import weakref

try:
    from ansys.mapdl.reader._reader import write_array

    _HAS_READER = True
except ModuleNotFoundError:  # pragma: no cover
    from ansys.mapdl.core.misc import write_array

    _HAS_READER = False

import numpy as np

from ansys.mapdl.core.mapdl import _MapdlCore
from ansys.mapdl.core.misc import supress_logging

ROUTINE_MAP = {
    0: "Begin level",
    17: "PREP7",
    21: "SOLUTION",
    31: "POST1",
    36: "POST26",
    52: "AUX2",
    53: "AUX3",
    62: "AUX12",
    65: "AUX15",
}

UNITS_MAP = {
    -1: "NONE",
    0: "USER",
    1: "SI",
    2: "CGS",
    3: "BFT",
    4: "BIN",
    5: "MKS",
    6: "MPA",
    7: "uMKS",
}


class Parameters:
    """Collection of MAPDL parameters.

    Notes
    -----

    See :ref:`ref_parameters` for additional notes.

    Examples
    --------
    Simply list all parameters except for MAPDL MATH parameters.

    >>> mapdl.parameters
    ARR                              : ARRAY DIM (3, 1, 1)
    PARM_FLOAT                       : 20.0
    PARM_INT                         : 10.0
    PARM_LONG_STR                    : "stringstringstringstringstringst"
    PARM_STR                         : "string"
    PORT                             : 50052.0

    Get a parameter

    >>> mapdl.parameters['PARM_FLOAT']
    20.0

    Get an array parameter

    >>> mapdl.parameters['ARR']
    array([1., 2., 3.])

    """

    def __init__(self, mapdl):
        if not isinstance(mapdl, _MapdlCore):
            raise TypeError("Must be implemented from MAPDL class")
        self._mapdl_weakref = weakref.ref(mapdl)
        self.show_leading_underscore_parameters = False
        self.show_trailing_underscore_parameters = False
        self.full_parameters_output = self._full_parameter_output(self)

    @property
    def _mapdl(self):
        """Return the weakly referenced instance of mapdl"""
        return self._mapdl_weakref()

    def _set_log_level(self, level):
        self._mapdl.set_log_level(level)

    @property
    def _log(self):
        return self._mapdl._log

    @property
    def numcpu(self) -> int:
        """Number of Distributed MAPDL processes being used.

        Notes
        -----
        This will always return ``1`` when using shared memory parallel.

        Examples
        --------
        >>> mapdl.parameters.numcpu
        2

        """
        return int(self._mapdl.get_value("ACTIVE", item1="NUMCPU"))

    @property
    def routine(self) -> str:
        """Current routine string as a string.  For example ``"/PREP7"``

        MAPDL Command: \*GET, ACTIVE, 0, ROUT

        Returns
        -------
        routine : str
            Routine as a string.  One of:

                - ``"Begin level"``
                - ``"PREP7"``
                - ``"SOLUTION"``
                - ``"POST1"``
                - ``"POST26"``
                - ``"AUX2"``
                - ``"AUX3"``
                - ``"AUX12"``
                - ``"AUX15"``

        Examples
        --------
        >>> mapdl.parameters.routine
        'PREP7'
        """
        value = self._mapdl.get_value("ACTIVE", item1="ROUT")
        return ROUTINE_MAP[int(value)]

    @property
    def units(self) -> str:
        """Units specified by /UNITS command.

        Returns
        -------
        units : str
            Active Units.  One of:
                - ``"None"``
                - ``"USER"``
                - ``"SI"``
                - ``"CGS"``
                - ``"BFT"``
                - ``"BIN"``
                - ``"MKS``
                - ``"MPA"``
                - ``"uMKS"``

        Examples
        --------
        >>> mapdl.parameters.units
        'NONE'
        """
        value = self._mapdl.get_value("ACTIVE", item1="UNITS")
        return UNITS_MAP[int(value)]

    @property
    def revision(self) -> float:
        """MAPDL revision version.

        Examples
        --------
        >>> mapdl.parameters.revision
        20.2
        """
        return float(self._mapdl.get_value("ACTIVE", item1="REV"))

    @property
    def platform(self) -> str:
        """The current platform.

        Examples
        --------
        >>> mapdl.parameters.platform
        'LIN'
        """
        return self._mapdl.get_value("ACTIVE", item1="PLATFORM")

    @property
    def csys(self) -> int:
        """Active coordinate system

        Examples
        --------
        >>> mapdl.parameters.csys
        0
        """
        return int(self._mapdl.get_value("ACTIVE", item1="CSYS"))

    @property
    def dsys(self) -> int:
        """Active display coordinate system

        Examples
        --------
        >>> mapdl.parameters.dsys
        0
        """
        return int(self._mapdl.get_value("ACTIVE", item1="DSYS"))

    @property
    def rsys(self) -> int:
        """Active result coordinate system

        Examples
        --------
        >>> mapdl.parameters.rsys
        0
        """
        return int(self._mapdl.get_value("ACTIVE", item1="RSYS"))

    @property
    def esys(self) -> int:
        """Active element coordinate system

        Examples
        --------
        >>> mapdl.parameters.esys
        0
        """
        return int(self._mapdl.get_value("ACTIVE", item1="ESYS"))

    @property
    def section(self) -> int:
        """Active section number

        Examples
        --------
        >>> mapdl.parameters.section
        1
        """
        return int(self._mapdl.get_value("ACTIVE", item1="SECT"))

    @property
    def material(self) -> int:
        """Active material

        Examples
        --------
        >>> mapdl.parameters.material
        1
        """
        return int(self._mapdl.get_value("ACTIVE", item1="MAT"))

    @property
    def real(self) -> int:
        """Active real constant set

        Examples
        --------
        >>> mapdl.parameters.real
        1
        """
        return int(self._mapdl.get_value("ACTIVE", item1="REAL"))

    @property
    def type(self) -> int:
        """Active element type

        Examples
        --------
        >>> mapdl.parameters.type
        1
        """
        return int(self._mapdl.get_value("ACTIVE", item1="type"))

    @property
    @supress_logging
    def _parm(self):
        """Current MAPDL parameters"""
        params = interp_star_status(self._mapdl.starstatus())

        if self.show_leading_underscore_parameters:
            _params = interp_star_status(self._mapdl.starstatus("_PRM"))
            params.update(_params)

        if self.show_trailing_underscore_parameters:
            params_ = interp_star_status(self._mapdl.starstatus("PRM_"))
            params.update(params_)

        return params

    def __repr__(self):
        """Return the current parameters in a pretty format"""
        lines = ["MAPDL Parameters", "----------------"]
        for key, info in self._parm.items():
            value_str = ""
            if info["type"] == "ARRAY":
                value_str = "ARRAY DIM %s" % str(info["shape"])
            elif info["type"] == "TABLE":
                value_str = "TABLE DIM %s" % str(info["shape"])
            elif info["type"] == "CHARACTER":
                value_str = '"%s"' % info["value"]
            elif "value" in info:
                value_str = str(info["value"])
            else:
                continue
            lines.append("%-32s : %s" % (key, value_str))
        return "\n".join(lines)

    def __getitem__(self, key):
        """Return a parameter"""
        if not isinstance(key, str):
            raise TypeError("Parameter name must be a string")
        key = key.upper()

        with self.full_parameters_output:
            parameters = self._parm

        if key not in parameters:
            raise IndexError("%s not a valid parameter_name" % key)

        parm = parameters[key]
        if parm["type"] in ["ARRAY", "TABLE"]:
            try:
                return self._get_parameter_array(key, parm["shape"])
            except ValueError:
                # allow a second attempt
                return self._get_parameter_array(key, parm["shape"])

        return parm["value"]

    def __setitem__(self, key, value):
        """Set a parameter"""
        self._mapdl._check_parameter_name(key)

        # parameters = self._parm  # check parameter exists
        if isinstance(value, (np.ndarray, list)):
            self._set_parameter_array(key, value)
        else:
            self._set_parameter(key, value)

    def __contains__(self, key):
        return key in self._parm.keys()

    def __iter__(self):
        yield from self._parm.keys()

    @supress_logging
    def _set_parameter(self, name, value):
        """Set a single parameter within MAPDL

        Parameters
        ----------
        name : str
            An alphanumeric name used to identify this parameter.  Name
            may be up to 32 characters, beginning with a letter and
            containing only letters, numbers, and underscores.
            Examples: ``"ABC" "A3X" "TOP_END"``.

        """
        if not isinstance(value, (str, int, float)):
            raise TypeError("``Parameter`` must be either a float, int, or string")

        if not isinstance(name, str):
            raise TypeError("``name`` must be a string")

        if len(name) > 32:
            raise ValueError("Length of ``name`` must be 32 characters or less")

        # delete the parameter if it exists as an array
        parm = self._parm
        if name in parm:
            if parm[name]["type"] == "ARRAY":
                self._mapdl.starset(name, mute=True)

        if isinstance(value, str):
            if " " in value:
                raise ValueError("Spaces not allowed in strings in MAPDL")
            self._mapdl.starset(name, f"'{value}'", mute=True)
        else:
            self._mapdl.starset(name, value, mute=True)

    @supress_logging
    def _get_parameter_array(self, parm_name, shape):
        """Return an ANSYS array parameter as a numpy.ndarray

        Parameters
        ----------
        parm_name : str
            MAPDL parameter name.

        Returns
        -------
        array : np.ndarray
            Numpy array.
        """
        escaped = False
        for each_format_number in [20, 30, 40, 64, 100]:
            format_str = f"(1F{each_format_number}.12)"
            with self._mapdl.non_interactive:
                # use C ordering
                self._mapdl.mwrite(parm_name.upper(), label="kji")
                self._mapdl.run(format_str)

            st = self._mapdl.last_response.rfind(format_str) + len(format_str) + 1

            if "**" not in self._mapdl.last_response[st:]:
                escaped = True
                break

        if not escaped:  # pragma: no cover
            raise RuntimeError(
                f"The array '{parm_name}' has a number format "
                "that could not be read using '{format_str}'."
            )

        arr_flat = np.fromstring(self._mapdl.last_response[st:], sep="\n").reshape(
            shape
        )

        if len(shape) == 3:
            if shape[2] == 1:
                arr_flat = arr_flat.squeeze(axis=2)

        return arr_flat

    @supress_logging
    def _set_parameter_array(self, name, arr):
        """Load a numpy array or python list directly to MAPDL

        Writes the numpy array to disk and then reads it in within
        MAPDL using \*VREAD.

        Parameters
        ----------
        arr : np.ndarray or List
            Array to send to MAPDL.  Maximum of 3 dimensions.

        name : str
            Name of the array to write to within MAPDL.

        Examples
        --------
        Load a 1D numpy array into MAPDL

        >>> arr = np.array([10, 20, 30])
        >>> mapdl.set_parameter_array(arr, 'MYARR')
        >>> parm, mapdl_arrays = mapdl.load_parameters()
        >>> mapdl_arrays['MYARR']
        array([10., 20., 30.])

        Load a 2D numpy array into MAPDL

        >>> arr = np.random.random((5, 3))
        >>> mapdl.set_parameter_array(arr, 'MYARR')
        >>> parm, mapdl_arrays = mapdl.load_parameters()
        >>> mapdl_arrays['MYARR']
        array([[0.39806635, 0.15060953, 0.3990557 ],
               [0.26837768, 0.02033222, 0.15655861],
               [0.46110226, 0.06381489, 0.20068533],
               [0.20122863, 0.5727896 , 0.85636037],
               [0.68126612, 0.67460878, 0.3678797 ]])

        Load a python list into MAPDL

        >>> mapdl.set_parameter_array([10, -1, 8, 4, 10], 'MYARR')
        >>> parm, mapdl_arrays = mapdl.load_parameters()
        >>> mapdl_arrays['MYARR']
        array([10., -1.,  8.,  4., 10.])

        """
        # type checks
        arr = np.array(arr)
        if not np.issubdtype(arr.dtype, np.number):
            raise TypeError("Only numerical arrays or lists are supported")
        if arr.ndim > 3:
            raise ValueError(
                "MAPDL VREAD only supports a arrays with a" " maximum of 3 dimensions."
            )

        if not isinstance(name, str):
            raise TypeError("``name`` must be a string")

        name = name.upper()

        idim, jdim, kdim = arr.shape[0], 1, 1
        if arr.ndim >= 2:
            jdim = arr.shape[1]
        if arr.ndim == 3:
            kdim = arr.shape[2]

        # 2021R1 and earlier has stability issues.  Directly stream
        # results for improved performance and stability, but up to a
        # size limit.
        if arr.size < 1000:
            return self._set_array_chain(name, arr, idim, jdim, kdim)

        # write array from numpy to disk
        filename = "_tmp.dat"
        self._write_numpy_array(filename, arr)

        cmd = f"{name}(1, 1),{filename},,,IJK,{idim},{jdim},{kdim}"
        with self._mapdl.non_interactive:
            self._mapdl.dim(name, imax=idim, jmax=jdim, kmax=kdim)
            self._mapdl.vread(cmd)
            self._mapdl.run("(1F20.12)")

    def _set_array_chain(self, name, arr, idim, jdim, kdim):
        """Sets an array using chained commands

        Faster and more stable (as of 2021R1) for small arrays.
        """
        if arr.ndim == 1:
            arr = np.expand_dims(arr, [1, 2])
        elif arr.ndim == 2:
            arr = np.expand_dims(arr, 2)

        # backwards compatibility with CORBA
        if hasattr(self._mapdl, "mute"):
            old_mute = self._mapdl.mute
            self._mapdl.mute = True

        with self._mapdl.non_interactive:
            self._mapdl.dim(name, imax=idim, jmax=jdim, kmax=kdim)
            for i in range(idim):
                for j in range(jdim):
                    for k in range(kdim):
                        index = f"{i + 1},{j + 1},{k + 1}"
                        self._mapdl.run(f"{name}({index})={arr[i, j, k]}")

        if hasattr(self._mapdl, "mute"):
            self._mapdl.mute = old_mute

    def _write_numpy_array(self, filename, arr):
        """Write a numpy array to disk"""
        if arr.dtype != np.double:
            arr = arr.astype(np.double)

        if self._mapdl._local:
            filename = os.path.join(self._mapdl.directory, filename)
        else:
            filename = os.path.join(tempfile.gettempdir(), filename)
        write_array(filename.encode(), arr.ravel("F"))

        if not self._mapdl._local:
            self._mapdl.upload(filename, progress_bar=False)

    class _full_parameter_output:
        """Change the show_** options to true to allow full parameter output."""

        def __init__(self, parent):
            self._parent = weakref.ref(parent)
            self.show_leading_underscore_parameters = None
            self.show_trailing_underscore_parameters = None

        def __enter__(self):
            """Storing current state."""
            self.show_leading_underscore_parameters = (
                self._parent().show_leading_underscore_parameters
            )
            self.show_trailing_underscore_parameters = (
                self._parent().show_trailing_underscore_parameters
            )

            # Getting full output.
            self._parent().show_leading_underscore_parameters = True
            self._parent().show_trailing_underscore_parameters = True

        def __exit__(self, *args):
            """Coming back to previous state."""
            self._parent().show_leading_underscore_parameters = (
                self.show_leading_underscore_parameters
            )
            self._parent().show_trailing_underscore_parameters = (
                self.show_trailing_underscore_parameters
            )


def interp_star_status(status):
    """Interprets \*STATUS command output from MAPDL

    Parameters
    ----------
    status : str
        Output from MAPDL *STATUS

    Returns
    -------
    parameters : dict
        Dictionary of parameters.
    """
    parameters = {}
    st = status.find("NAME                              VALUE")

    if st == -1:
        return {}

    for line in status[st + 80 :].splitlines():
        items = line.split()
        if not items:
            continue

        # line will contain either a character, scalar, or array
        name = items[0]
        if len(items) == 2:
            if items[1][-9:] == "CHARACTER":
                parameters[name] = {"type": "CHARACTER", "value": items[1][:-9]}
            # else:
            # log.warning(
        elif len(items) == 3:
            if items[2] == "SCALAR":
                value = float(items[1])
            else:
                value = items[1]
            parameters[name] = {"type": items[2], "value": value}
        elif len(items) == 5:
            shape = (int(items[2]), int(items[3]), int(items[4]))
            parameters[name] = {"type": items[1], "shape": shape}
    return parameters



"""
These PREP7 commands define the model real constants.
"""


class RealConstants:
    def r(self, nset="", r1="", r2="", r3="", r4="", r5="", r6="", **kwargs):
        """APDL Command: R

        Defines the element real constants.

        Parameters
        ----------
        nset
            Real constant set identification number (arbitrary).  If same as a
            previous set number, set is redefined. Set number relates to that
            defined with the element [REAL]. Note that the GUI automatically
            assigns this value.

        r1, r2, r3, . . . , r6
            Real constant values (interpreted as area, moment of inertia,
            thickness, etc., as required for the particular element type using
            this set), or table names for tabular input of boundary conditions.
            Use RMORE command if more than six real constants per set are to be
            input.

        Notes
        -----
        Defines the element real constants.  The real constants required for an
        element are shown in the Input Summary of each element description in
        the Element Reference.  Constants must be input in the same order as
        shown in that table.  If more than the required number of element real
        constants are specified in a set, only those required are used.  If
        fewer than the required number are specified, zero values are assumed
        for the unspecified constants.

        If using table inputs (SURF151, SURF152, FLUID116, CONTA171, CONTA172,
        CONTA173, CONTA174, and CONTA175 only), enclose the table name in %
        signs (e.g., %tabname%).

        Specify NSET = GCN to define real constants for real constant sets that
        were previously assigned by the GCDEF command (that is, real constants
        used in general contact interactions).

        When copying real constants to new sets, ANSYS, Inc. recommends that
        you use the command input. If you do use the GUI, restrict the real
        constant copy to only the first six real constants (real constants
        seven and greater will be incorrect for both the master and copy set).

        This command is also valid in SOLUTION.
        """
        command = "R,%s,%s,%s,%s,%s,%s,%s" % (
            str(nset),
            str(r1),
            str(r2),
            str(r3),
            str(r4),
            str(r5),
            str(r6),
        )
        return self.run(command, **kwargs)

    def rdele(self, nset1="", nset2="", ninc="", lchk="", **kwargs):
        """APDL Command: RDELE

        Deletes real constant sets.

        Parameters
        ----------
        nset1, nset2, ninc
            Delete real constant sets from NSET1 to NSET2 (defaults to NSET1)
            in steps of NINC (defaults to 1).  If NSET1 = ALL, ignore NSET2 and
            NINC and all real constant sets are deleted.

        lchk
            Specifies the level of element-associativity checking:

            NOCHECK - No element-associativity check occurs. This option is the default.

            WARN - When a section, material, or real constant is associated with an element, ANSYS
                   issues a message warning that the necessary entity has been
                   deleted.

            CHECK - The command terminates, and no section, material, or real constant is deleted
                    if it is associated with an element.

        Notes
        -----
        Deletes real constant sets defined with the R command.

        This command is also valid in SOLUTION.
        """
        command = "RDELE,%s,%s,%s,%s" % (str(nset1), str(nset2), str(ninc), str(lchk))
        return self.run(command, **kwargs)

    def rlist(self, nset1="", nset2="", ninc="", **kwargs):
        """APDL Command: RLIST

        Lists the real constant sets.

        Parameters
        ----------
        nset1, nset2, ninc
            List real constant sets from NSET1 to NSET2 (defaults to NSET1) in
            steps of NINC (defaults to 1).  If NSET1 = ALL (default), ignore
            NSET2  and NINC and list all real constant sets [R].

        Notes
        -----
        The real constant sets listed contain only those values specifically
        set by the user.  Default values for real constants set automatically
        within the various elements are not listed.

        This command is valid in any processor.
        """
        command = "RLIST,%s,%s,%s" % (str(nset1), str(nset2), str(ninc))
        return self.run(command, **kwargs)

    def rmodif(
        self, nset="", stloc="", v1="", v2="", v3="", v4="", v5="", v6="", **kwargs
    ):
        """APDL Command: RMODIF

        Modifies real constant sets.

        Parameters
        ----------
        nset
            Number of existing real constant set to be modified.

        stloc
            Starting location in table for modifying data.  For example, if
            STLOC = 1, data input in the V1 field is the first constant in the
            set.  If STLOC = 7, data input in the V1 field is the seventh
            constant in the set, etc.  Must be greater than zero.

        v1
            New value assigned to constant in location STLOC.  If zero (or
            blank), a zero value will be assigned.

        v2, v3, v4, . . . , v6
            New values assigned to constants in the next five locations.  If
            blank, the value remains unchanged.

        Notes
        -----
        Allows modifying (or adding) real constants to an existing set [R] at
        any location.

        Specify NSET = GCN to define/modify real constants for real constant
        sets that were previously assigned by the GCDEF command (that is, real
        constants used in general contact interactions).

        This command is also valid in SOLUTION. For important information about
        using this command within the solution phase, see What Are Nonstandard
        Uses? in the Advanced Analysis Guide.
        """
        command = "RMODIF,%s,%s,%s,%s,%s,%s,%s,%s" % (
            str(nset),
            str(stloc),
            str(v1),
            str(v2),
            str(v3),
            str(v4),
            str(v5),
            str(v6),
        )
        return self.run(command, **kwargs)

    def rmore(self, r7="", r8="", r9="", r10="", r11="", r12="", **kwargs):
        """APDL Command: RMORE

        Adds real constants to a set.

        Parameters
        ----------
        r7, r8, r9, . . . , r12
            Add real constants 7 to 12 (numerical values or table names) to the
            most recently defined set.

        Notes
        -----
        Adds six more real constants to the most recently defined set.  Repeat
        the RMORE command for constants 13 to 18, again for 19-24, etc.

        If using table inputs (SURF151, SURF152, FLUID116, CONTA171, CONTA172,
        CONTA173, CONTA174, and CONTA175 only), enclose the table name in %
        signs (e.g., %tabname%).

        When copying real constants to new sets, ANSYS, Inc. recommends that
        you use the command input. If you do use the GUI, restrict the real
        constant copy to only the first six real constants (real constants
        seven and greater will be incorrect for both the master and copy set).

        This command is also valid in SOLUTION.
        """
        command = "RMORE,%s,%s,%s,%s,%s,%s" % (
            str(r7),
            str(r8),
            str(r9),
            str(r10),
            str(r11),
            str(r12),
        )
        return self.run(command, **kwargs)

    def setfgap(
        self, gap="", ropt="", pamb="", acf1="", acf2="", pref="", mfp="", **kwargs
    ):
        """APDL Command: SETFGAP

        Updates or defines the real constant table for squeeze film elements.

        Parameters
        ----------
        gap
            Gap separation.

        ropt
            Real constant set option.

            0 - Creates separate real constant sets for each selected element with the
                specified real constant values (default).

            1 - Updates existing real constant sets. The gap separation is updated from
                displacement results in the database. Other real constants are
                updated as specified in the command input parameters.

        pamb
            Ambient pressure.

        acf1, acf2
            Accommodation factor 1 and 2.

        pref
            Reference pressure for mean free path.

        mfp
            Mean free path.

        Notes
        -----
        This command is used for large signal cases to update the gap
        separation real constant on a per-element basis.  Issue this command
        prior to solution using the default ROPT value to initialize real
        constant sets for every fluid element.  After a solution, you can re-
        issue the command to update the real constant set for a subsequent
        analysis. See Introduction for more information on thin film analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"SETFGAP,{gap},{ropt},,{pamb},{acf1},{acf2},{pref},{mfp}"
        return self.run(command, **kwargs)



class ListControls:
    def com(self, comment="", **kwargs):
        """Places a comment in the output.

        APDL Command: /COM

        Parameters
        ----------
        comment
            Comment string, up to 75 characters.

        Notes
        -----
        The output from this command consists of the comment string.  This
        command is similar to C*** except that the comment produced by C*** is
        more easily identified in the output. Parameter substitution within the
        comment occurs for every valid expression delimited by percent (%)
        signs. Enclosing such an expression in single quotes prevents parameter
        substitution.

        Another way to include a comment is to precede it with a ! character
        (on the same line).  The ! may be placed anywhere on the line, and any
        input following it is ignored as a comment.  No output is produced by
        such a comment, but the comment line is included on the log file.  This
        is a convenient way to annotate the log file.

        This command is valid anywhere.
        """
        command = "/COM,%s" % (str(comment))
        if self.print_com and not self.mute and not kwargs.get("mute", False):
            print(command)
        return self.run(command, **kwargs)

    def golist(self, **kwargs):
        """Reactivates the suppressed data input listing.

        APDL Command: /GOLIST

        Notes
        -----
        Reactivates printout of the data input listing suppressed with /NOLIST.

        This command is valid in any processor, but only within a batch run
        [/BATCH].
        """
        command = "/GOLIST,"
        return self.run(command, **kwargs)

    def gopr(self, **kwargs):
        """Reactivates suppressed printout.

        APDL Command: /GOPR

        Notes
        -----
        Reactivates printout suppressed with the /NOPR command.  The /GO
        command has the same function except that it does not produce a command
        response from the program.

        This command is valid in any processor.
        """
        command = "/GOPR,"
        return self.run(command, **kwargs)

    def nolist(self, **kwargs):
        """Suppresses the data input listing.

        APDL Command: /NOLIST

        Notes
        -----
        Printout is suppressed until a /GOLIST command is read or the end of
        the listing is encountered.

        This command is valid in any processor, but only within a batch run
        [/BATCH].
        """
        command = "/NOLIST,"
        return self.run(command, **kwargs)

    def nopr(self, **kwargs):
        """Suppresses the expanded interpreted input data listing.

        APDL Command: /NOPR

        Notes
        -----
        Suppresses printout of interpreted input data, including information
        labeled as "Notes."  When this printout is not suppressed, the data
        input to the analysis is echoed to the output file in an expanded
        format.  Printout is suppressed until a /GOPR or /GO command is read.

        Use of /NOPR is not recommended when the graphical user interface (GUI)
        is active.  The GUI sometimes issues "hidden" /NOPR and /GOPR command
        sequences, which will countermand user-issued /NOPR commands, thus
        making the use of /NOPR in the GUI environment unpredictable.

        This command is valid in any processor.
        """
        command = "/NOPR,"
        return self.run(command, **kwargs)



class Controls:
    def cfact(
        self, rfacta="", ifacta="", rfactb="", ifactb="", rfactc="", ifactc="", **kwargs
    ):
        """Defines complex scaling factors to be used with operations.

        APDL Command: CFACT

        Parameters
        ----------
        rfacta
            Real portion of the complex scale factor used in place of FACTA.

        ifacta
            Imaginary portion of the complex scale factor used in place of
            FACTA.

        rfactb
            Real portion of the complex scale factor used in place of FACTB.

        ifactb
            Imaginary portion of the complex scale factor used in place of
            FACTB.

        rfactc
            Real portion of the complex scale factor used in place of FACTC.

        ifactc
            Imaginary portion of the complex scale factor used in place of
            FACTC.

        Notes
        -----
        Defines complex scale factors to be used with the operations [ADD,
        PROD, etc.].  If this command is supplied, these complex factors
        override any real factors (FACTA, FACTB, FACTC) supplied on the
        operation commands.  Factors are typically involved in scaling a
        specified variable, such as in the term FACTA x IA of the ADD command
        to scale variable IA before the ADD operation.

        When the CFACT command is active, defaults are as follows: 1) if the
        complex factor is not specified, but the variable upon which it acts
        (such as IA) is specified, the factor defaults to 1.0+i0.0;  2) if the
        variable upon which the factor operates is not specified, but the
        factor is specified, the variable defaults to 1.0 so that the term in
        the operation becomes the complex factor itself;  3) if neither the
        factor nor the variable number is supplied, the term is omitted from
        the operation.  Once the operation (such as the ADD command) has been
        processed, the CFACT command becomes inactive and must be specified
        again if it is to be used.
        """
        command = f"CFACT,{rfacta},{ifacta},{rfactb},{ifactb},{rfactc},{ifactc}"
        return self.run(command, **kwargs)

    def force(self, lab="", **kwargs):
        """Selects the element nodal force type for output.

        APDL Command: FORCE

        Parameters
        ----------
        lab
            Type of force to be associated with the force items:

            TOTAL - Total forces (static, damping, and inertia).

            STATIC - Static forces.

            DAMP - Damping forces.

            INERT - Inertia forces.

        Notes
        -----
        FORCE selects the element nodal force type for output with the POST1
        PRESOL, PLESOL, PRRFOR, NFORCE, FSUM, etc. commands, the POST26 ESOL
        command, and reaction force plotting [/PBC].  For example, FORCE,STATIC
        causes item F of the PRESOL command to be the static forces for the
        elements processed. Element member forces (such as those available for
        beams and shells and processed by Item and Sequence number) are not
        affected by this command. The SMISC records extract the static force.

        The PRRSOL command is not valid with FORCE.  Use the PRRFOR command,
        which provides the same functionality as PRRSOL, instead.

        Use the FORCE command prior to any load case operations (LCOPER) to
        insure the correct element nodal force combinations.

        In POST26, the ESOL data stored is based on the active FORCE
        specification at the time the data is stored. To store data at various
        specifications (for example, static and inertia forces), issue a STORE
        command before each new specification.

        The FORCE command cannot be used to extract static, damping, and
        inertial forces for MPC184 joint elements.

        To retrieve the different force types, use the ``*GET`` command with
        Entity=ELEM and Item1=EFOR.

        The FORCE command is not supported in a spectrum analysis. You can
        specify the force type directly on the combination method commands
        (ForceType on the PSDCOM, SRSS, CQC, etc. commands).

        The FORCE command is not supported in a modal analysis.
        """
        command = f"FORCE,{lab}"
        return self.run(command, **kwargs)

    def layerp26(self, num="", **kwargs):
        """Specifies the element layer for which data are to be stored.

        APDL Command: LAYERP26

        Parameters
        ----------
        num
            Layer-processing mode:

            N - The layer number to process. The default value is 1.

        Notes
        -----
        Defines the element layer for which results data are to be stored for
        postprocessing.  Applies to stress and strain data for layered elements
        BEAM161, SHELL163, SHELL181, SOLID185, SOLID186, SOLSH190, SHELL208,
        SHELL209, SHELL281, REINF265, and ELBOW290.

        The SHELL command can be used (for shell elements) to specify a
        location (TOP, MID, BOT) within the layer for selection on the ESOL
        command. Transverse shear stresses for MID are linearly averaged from
        TOP and BOT, and do not reflect a parabolic distribution. Setting
        KEYOPT(8) = 2 for SHELL181, SHELL208, SHELL209, SHELL281, and ELBOW290
        writes the mid-surface values directly to the results file and yields
        more accurate values than linear averaging.

        That this command cannot be used for energy output, as energy is a per-
        element quantity.

        When using the LAYERP26 command with SHELL181, SOLID185, SOLID186,
        SOLSH190, SHELL208, or SHELL209, KEYOPT(8) must be set to 1 (or 2 for
        SHELL181, SHELL208, SHELL209, SHELL281, and ELBOW290) in order to store
        results for all layers.

        For the ANSYS LS-DYNA product, this command works differently than
        described above.  For SHELL163 and BEAM161, you must first use EDINT
        during the solution phase to define the integration points for which
        you want output data.  Be aware that the output location for SHELL163
        data is always at the integration point, so "top" and "bottom" refer to
        the top or bottom integration point, not necessarily the top or bottom
        surface.  For more information, see the ANSYS LS-DYNA User's Guide.

        In POST26, the ESOL data stored is based on the active LAYERP26
        specification at the time the data is stored. To store data at various
        specifications (for example, layers 2 and 5), issue a STORE command
        before each new specification.
        """
        command = f"LAYERP26,{num}"
        return self.run(command, **kwargs)

    def shell(self, loc="", **kwargs):
        """Selects a shell element or shell layer location for results output.

        APDL Command: SHELL

        Parameters
        ----------
        loc
            Location within shell element (or layer) to obtain stress results:

            TOP - Top of shell element (or layer) (default).

            MID - Middle of shell element (or layer). The default method averages the TOP and BOT
                  values to obtain a mid value. Setting KEYOPT(8) = 2 for
                  SHELL181, SHELL208, SHELL209, and ELBOW290 uses MID results
                  obtained directly from the results file.

            BOT - Bottom of shell element (or layer).

        Notes
        -----
        Selects the location within a shell element (or a shell layer) for
        results output (nodal stresses, strains, etc.).  Applies to POST1
        selects, sorts, and output [NSEL, NSORT, PRNSOL, PLNSOL, PRPATH,
        PLPATH, etc.], and is used for storage with the POST26 ESOL command.
        For example, SHELL,TOP causes item S of the POST1 PRNSOL command or the
        POST26 ESOL command to be the stresses at the top of the shell
        elements.  For layered shell elements, use the LAYER (POST1) or
        LAYERP26 (POST26) command to select the layer. The SHELL command does
        not apply to the layered thermal shell elements, SHELL131 and SHELL132.

        For PowerGraphics [/GRAPHICS,POWER], the SHELL,MID command affects both
        the printed output and the displayed results, while the SHELL (TOP or
        BOT) command prints and displays both the top and bottom layers
        simultaneously. Note that /CYCEXPAND,ON automatically turns on
        PowerGraphics; however, for cyclic mode-superposition harmonic
        postprocessing (CYCFILES), the SHELL command prints and displays only
        the requested layer.

        In POST26, the ESOL data stored is based on the active SHELL
        specification at the time the data is stored. To store data at various
        specifications (for example, stresses at the top and bottom locations),
        issue a STORE command before each new specification.
        """
        command = f"SHELL,{loc}"
        return self.run(command, **kwargs)

    def tvar(self, key="", **kwargs):
        """Changes time to the cumulative iteration number.

        APDL Command: TVAR

        Parameters
        ----------
        key
            Time key:

            0 - Time is used for the variable TIME.

            1 - NCUMIT is used for the variable TIME.

        Notes
        -----
        Changes the meaning of the time variable to the cumulative iteration
        number (NCUMIT) variable.  Data can be read from the file, printed, and
        displayed as a function of NCUMIT rather than time.  All POST26
        descriptions applying to TIME then apply to NCUMIT.
        """
        command = f"TVAR,{key}"
        return self.run(command, **kwargs)



class Scaling:
    def iclwid(self, factor="", **kwargs):
        """Scales the line width of circuit builder icons.

        APDL Command: /ICLWID

        Parameters
        ----------
        factor
            Multiplication factor applied to the default line width (defaults
            to 1). The minimum is 1 and the maximum is 6.

        Notes
        -----
        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"/ICLWID,{factor}"
        return self.run(command, **kwargs)

    def icscale(self, wn="", factor="", **kwargs):
        """Scales the icon size for elements supported in the circuit builder.

        APDL Command: /ICSCALE

        Parameters
        ----------
        wn
            Window number (or ALL) to which command applies (defaults to 1).

        factor
            Factor applied to the default icon size (defaults to 1).

        Notes
        -----
        Scaling the icon size can provide better visualization of the circuit
        components when using the Circuit Builder (an interactive builder
        available in the ANSYS GUI).

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"/ICSCALE,{wn},{factor}"
        return self.run(command, **kwargs)

    def ratio(self, wn="", ratox="", ratoy="", **kwargs):
        """Distorts the object geometry.

        APDL Command: /RATIO

        Parameters
        ----------
        wn
            Window number (or ALL) to which command applies (defaults to 1).

        ratox
            Distort object in the window X direction by this factor (defaults
            to 1.0).

        ratoy
            Distort object in the window Y direction by this factor (defaults
            to 1.0).

        Notes
        -----
        Distorts the object geometry in a particular direction.  An example of
        this command's use would be to allow long narrow sections to be
        distorted to a more square area for better display visualization.

        This command is valid in any processor.
        """
        command = f"/RATIO,{wn},{ratox},{ratoy}"
        return self.run(command, **kwargs)

    def shrink(self, ratio="", **kwargs):
        """Shrinks elements, lines, areas, and volumes for display clarity.

        APDL Command: /SHRINK

        Parameters
        ----------
        ratio
            Shrinkage ratio (input as a decimal (0.0 to 0.5)).  Defaults to 0.0
            (no shrinkage).  Values greater than 0.5 default to 0.1 (10%
            shrinkage).

        Notes
        -----
        Shrinks the elements, lines, areas, and volumes so that adjacent
        entities are separated for clarity.  Portions of this command are not
        supported by PowerGraphics [/GRAPHICS,POWER].

        If only the common lines of non-coplanar faces are drawn (as per the
        /EDGE command), then this command is ignored.

        This command is valid in any processor.
        """
        command = f"/SHRINK,{ratio}"
        return self.run(command, **kwargs)

    def sscale(self, wn="", smult="", **kwargs):
        """Sets the contour multiplier for topographic displays.

        APDL Command: /SSCALE

        Parameters
        ----------
        wn
            Window number (or ALL) to which command applies (defaults to 1).

        smult
            Contour multiplier that factors in results based on the product of
            the multiplier and the result being plotted. Defaults to 0.0 (no
            topographic effects).

        Notes
        -----
        Use this command to scale values to the geometry when the contours are
        shown elevated.  For section displays [/TYPE], the elevation is
        performed perpendicular to the section face.

        Nonzero contour multipliers factoring in large results (stresses or
        displacements) can produce very large distortion, causing images to
        disappear.  To bring a distorted image back into view, reduce the
        contour multiplier value.

        Portions of this command are not supported by PowerGraphics
        [/GRAPHICS,POWER].
        """
        command = f"/SSCALE,{wn},{smult}"
        return self.run(command, **kwargs)

    def txtre(self, lab="", num="", n1="", n2="", ninc="", **kwargs):
        """Controls application of texture to selected items.

        APDL Command: /TXTRE

        Parameters
        ----------
        lab
            You can apply texture according to the following labels:

            ELEM - Apply texture to elements N1 through N2 in steps of NINC.

            AREA - Apply texture to areas N1 through N2 in steps of NINC.

            VOLU - Apply texture to volumes N1 through N2 in steps of NINC.

            CM - Apply texture to the component named in N1. N2 and
            NINC are ignored.

            ON, OFF - Sets the specified texture display on or
            off. All other fields are ignored.

            File - If Lab = File, the command format is /TXTRE, File,
                   Key_Index, Fname, Fext, --, Format (This variant of
                   the command is applicable to 2-D drivers).

            Key_Index - The texture index associated with the file. If
                        the number fifty-one (51) is used, the
                        imported bitmap will be used as the window's
                        logo.

            Fname - File name and directory path (248 characters
                    maximum, including the characters needed for the
                    directory path).  An unspecified directory path
                    defaults to the working directory; in this case,
                    you can use all 248 characters for the file name.

            Fext - Filename extension (eight-character maximum).

            Format - The file format. If Format = 0, the file is a
                     pixmap (Linux) or Bitmap (PC).  The file cannot
                     contain a compressed image, and the PC file must
                     be 8 or 24 bit BI_RGB format. If Format = 1 or
                     JPEG, then the file is in JPEG (Joint
                     Photographic Experts Group) format. If Format = 2
                     or PNG, then the file is in PNG (Portable Network
                     Graphics) format.

        num
            Select the texture index number from the following list:

            0 - No Texturing

            1 - Aluminum

            2 -  Aluminum, Brushed

            3 - Steel With Bumps

            4 - Steel, Embossed

            5 - Iron

            6 - Steel, Pattern

            7 - Steel, Riveted

            8 - Steel, Scratched

            9 - Tin

            10 - Metal

            11 - Steel, Etched

            12 - Metal, Hot

            13 - Iron, Grainy

            14 - Metal, Rusty

            15 - Brick

            16 - Block

            17 - Wood

            18 - Wood, Light

            19 - Wood, Walnut

            20 - Plastic, Hard Blue

            21 - Plastic, Light Blue

            22 - Plastic, Hard Red

            31 - Gold

            32 - Brass

            33 - Silver

            34 -  Plastic, Black

            35 - Plastic, Ivory

            36 - Plastic, Blue

            37 - Plastic, Red

            38 - Plastic, Yellow

            39 - Plastic, Green

            40 - Plastic, Brown

        n1, n2, ninc
            Apply texture to Lab items numbered N1 through N2 in steps of NINC
            (defaults to 1). If Lab = CM, then N1 is used to for the component
            name and N2 and NINC are ignored. If Lab = ELEM, AREA, or VOLU and
            N1 = blank or ALL, then the specified texture will be applied to
            all entities of type Lab. If N1 = P, then graphical picking is
            enabled.

        Notes
        -----
        This command is available for 3-D Open GL devices. 2-D devices are
        supported only for the Lab = File variation of the command, allowing
        imported bitmaps to be used for texturing and annotation. Textures can
        affect the speed of many of your display operations. You can increase
        the speed by temporarily turning the textures off (Utility Menu>
        PlotCtrls> Style> Texturing(3D)> Display Texturing). This menu
        selection toggles your textures on and off. When textures are toggled
        off, all of the texture information is retained and reapplied when
        texturing is toggled back on.

        For some displays, the texture will appear distorted because of a
        technique used to enhance 3-D displays (/DV3D,TRIS,1). Disabling this
        function (/DV3D,TRIS,0) will improve the quality of some texture
        displays. Disabling the TRIS option of the /DV3D command will slow down
        3-D displays significantly. Be sure to reapply the TRIS option after
        you obtain a satisfactory output.

        Specifying /TXTRE,DEFA removes all texturing.
        """
        command = f"/TXTRE,{lab},{num},{n1},{n2},{ninc}"
        return self.run(command, **kwargs)

    def vscale(self, wn="", vratio="", key="", **kwargs):
        """Scales the length of displayed vectors.

        APDL Command: /VSCALE

        Parameters
        ----------
        wn
            Window number (or ALL) to which command applies (defaults to 1).

        vratio
            Ratio value applied to the automatically calculated scale factor
            (defaults to 1.0, i.e., use scale factor as automatically
            calculated).

        key
            Relative scaling key:

            0 - Use relative length scaling among vectors based on magnitudes.

            1 - Use uniform length scaling for all vector lengths.

        Notes
        -----
        Allows scaling of the vector length displayed with the PLVECT command
        of POST1 and the /PBC and /PSF commands.  Also allows the scaling of
        the  element (i.e., /PSYMB,ESYS) and the nodal (i.e., /PSYMB,NDIR)
        coordinate system symbols.

        This command is valid in any processor.
        """
        command = f"/VSCALE,{wn},{vratio},{key}"
        return self.run(command, **kwargs)



"""Module to control interaction with an ANSYS shell instance.

Used when launching Mapdl via pexpect on Linux when <= 17.0
"""
import os
import re
import time

from ansys.mapdl.core.errors import MapdlExitedError

# from ansys.mapdl.core.misc import kill_process
from ansys.mapdl.core.mapdl import _MapdlCore

ready_items = [
    rb"BEGIN:",
    rb"PREP7:",
    rb"SOLU_LS[0-9]+:",
    rb"POST1:",
    rb"POST26:",
    rb"RUNSTAT:",
    rb"AUX2:",
    rb"AUX3:",
    rb"AUX12:",
    rb"AUX15:",
    # continue
    rb"YES,NO OR CONTINUOUS\)\=",
    rb"executed\?",
    # errors
    rb"SHOULD INPUT PROCESSING BE SUSPENDED\?",
    # prompts
    rb"ENTER FORMAT for",
]

CONTINUE_IDX = ready_items.index(rb"YES,NO OR CONTINUOUS\)\=")
WARNING_IDX = ready_items.index(rb"executed\?")
ERROR_IDX = ready_items.index(rb"SHOULD INPUT PROCESSING BE SUSPENDED\?")
PROMPT_IDX = ready_items.index(rb"ENTER FORMAT for")

nitems = len(ready_items)
expect_list = []
for item in ready_items:
    expect_list.append(re.compile(item))
ignored = re.compile(r"[\s\S]+".join(["WARNING", "command", "ignored"]))


def launch_pexpect(
    exec_file=None,
    run_location=None,
    jobname=None,
    nproc=None,
    additional_switches="",
    start_timeout=60,
):
    """Launch MAPDL as a pexpect process.

    Limited to only a linux instance
    """
    import pexpect

    command = "%s -j %s -np %d %s" % (exec_file, jobname, nproc, additional_switches)
    process = pexpect.spawn(command, cwd=run_location)
    process.delaybeforesend = None

    try:
        index = process.expect(["BEGIN:", "CONTINUE"], timeout=start_timeout)
    except:  # capture failure
        raise RuntimeError(process.before.decode("utf-8"))

    if index:  # received ... press enter to continue
        process.sendline("")
        process.expect("BEGIN:", timeout=start_timeout)

    return process


class MapdlConsole(_MapdlCore):
    """Control interaction with an ANSYS shell instance.

    Only works on Linux.
    """

    def __init__(
        self,
        loglevel="INFO",
        log_apdl=None,
        use_vtk=True,
        print_com=False,
        **start_parm,
    ):
        """Opens an ANSYS process using pexpect"""
        self._auto_continue = True
        self._continue_on_error = False
        self._process = None
        self._launch(start_parm)
        super().__init__(
            loglevel=loglevel,
            use_vtk=use_vtk,
            log_apdl=log_apdl,
            print_com=print_com,
            **start_parm,
        )

    def _launch(self, start_parm):
        """Connect to MAPDL process using pexpect"""
        self._process = launch_pexpect(**start_parm)

    def _run(self, command, **kwargs):
        """Sends command and returns ANSYS's response"""
        self._reset_cache()

        if not self._process.isalive():
            raise MapdlExitedError("ANSYS exited")

        command = command.strip()
        if not command:
            raise ValueError("Cannot run empty command")

        if command[:4].lower() == "/out":
            items = command.split(",")
            if len(items) > 1:
                self._output = ".".join(items[1:])
            else:
                self._output = ""

        # send the command
        self._log.debug("Sending command %s", command)
        self._process.sendline(command)

        # do not expect
        if "/MENU" in command:
            self._log.info("Enabling GUI")
            self._process.sendline(command)
            return

        full_response = ""
        while True:
            i = self._process.expect_list(expect_list, timeout=None)
            response = self._process.before.decode("utf-8")
            full_response += response
            if i >= CONTINUE_IDX and i < WARNING_IDX:  # continue
                self._log.debug(
                    "Continue: Response index %i.  Matched %s",
                    i,
                    ready_items[i].decode("utf-8"),
                )
                self._log.info(response + ready_items[i].decode("utf-8"))
                if self._auto_continue:
                    user_input = "y"
                else:
                    user_input = input("Response: ")
                self._process.sendline(user_input)

            elif i >= WARNING_IDX and i < ERROR_IDX:  # warning
                self._log.debug(
                    "Prompt: Response index %i.  Matched %s",
                    i,
                    ready_items[i].decode("utf-8"),
                )
                self._log.warning(response + ready_items[i].decode("utf-8"))
                if self._auto_continue:
                    user_input = "y"
                else:
                    user_input = input("Response: ")
                self._process.sendline(user_input)

            elif i >= ERROR_IDX and i < PROMPT_IDX:  # error
                self._log.debug(
                    "Error index %i.  Matched %s", i, ready_items[i].decode("utf-8")
                )
                self._log.error(response)
                response += ready_items[i].decode("utf-8")
                raise Exception(response)

            elif i >= PROMPT_IDX:  # prompt
                self._log.debug(
                    "Prompt index %i.  Matched %s", i, ready_items[i].decode("utf-8")
                )
                self._log.info(response + ready_items[i].decode("utf-8"))
                raise RuntimeError(
                    "User input expected.  " "Try using ``with mapdl.non_interactive``"
                )
            else:  # continue item
                self._log.debug(
                    "continue index %i.  Matched %s", i, ready_items[i].decode("utf-8")
                )
                break

        # return last response and all preceding responses
        return full_response

    def exit(self, close_log=True, timeout=3):
        """Exit MAPDL process.

        Parameters
        ----------
        timeout : float
            Maximum time to wait for MAPDL to exit.  Set to 0 or
            ``None`` to not wait until MAPDL stops.
        """
        self._log.debug("Exiting ANSYS")
        if self._process is not None:
            try:
                self._process.sendline("FINISH")
                self._process.sendline("EXIT")
            except:
                pass

        if close_log:
            self._close_apdl_log()

        self._exited = True

        # edge case: need to wait until process dies, otherwise future
        # commands might talk to a dead process...
        if timeout:
            tstart = time.time()
            while self._process.isalive():
                time.sleep(0.05)
                telap = tstart - time.time()
                if telap > timeout:
                    return 1

        return 0

    def kill(self):
        """Forces ANSYS process to end and removes lock file"""
        if self._process is not None:
            try:
                self.exit()
            except:
                try:
                    os.kill(self._process.pid, 9)
                except:
                    self._log.warning("Unable to kill process %d", self._process.pid)
                self._log.debug("Killed process %d", self._process.pid)

    @property
    def _name(self):
        """Instance unique identifier."""
        return f"Console_PID_{self._process.pid}"



class Controls:
    def avprin(self, key="", effnu="", **kwargs):
        """Specifies how principal and vector sums are to be calculated.

        APDL Command: AVPRIN

        Parameters
        ----------
        key
            Averaging key:

            0 - Average the component values from the elements at a common node, then calculate
                the principal or vector sum from the averaged components
                (default).

            1 - Calculate the principal or vector sum values on a per element basis, then
                average these values from the elements at a common node.

        effnu
            Effective Poisson's ratio used for computing the von Mises
            equivalent strain (EQV). This command option is intended for use
            with line elements or in load case operations (LCOPER) only; ANSYS
            automatically selects the most appropriate effective Poisson's
            ratio, as discussed below.

        Notes
        -----
        Selects the method of combining components for certain derived nodal
        results when two or more elements connect to a common node.  The
        methods apply to the calculations of derived nodal principal stresses,
        principal strains, and vector sums for selects, sorts, and output
        [NSEL, NSORT, PRNSOL, PLNSOL, etc.].

        This command also defines the effective Poisson's ratio (EFFNU) used
        for equivalent strain calculations.  If you use EFFNU, the default
        effective Poisson's ratios shown below will be overridden for all
        elements by the EFFNU value. To return to the default settings, issue
        the RESET command. The default value for EFFNU is:

        Poisson's ratio as defined on the MP commands for EPEL and EPTH

        0.5 for EPPL and EPCR

        0.5 if the referenced material is hyperelastic

        0.0 for line elements (includes beam, link, and pipe elements, as well
        as discrete elements), cyclic symmetry analysis, mode superposition
        analyses (with MSUPkey = YES on the MXPAND command), and load case
        operations (LCOPER).

        For the von Mises equivalent strain (EQV), it is always computed using
        the average of the equivalent strains from the elements at a common
        node irrespective of the value of the averaging KEY. If EFFNU is input,
        though, the calculation will be performed according to the KEY setting.

        For a random vibration (PSD) analysis, issuing either AVPRIN,0 or
        AVPRIN,1 calculates the principal stresses using the appropriate
        averaging method. They are then used to determine SEQV.  The output
        will have non-zero values for the principal stresses.

        If AVPRIN is not issued, the Segalman-Fulcher method is used to
        calculate SEQV. This method does not calculate principal stresses, but
        directly calculates SEQV from the component stresses; therefore, the
        output will have zero values for the principal stresses. Beam and pipe
        elements are excluded

        This command is also valid in POST26, where applicable.

        See Combined Stresses and Strains in the Mechanical APDL Theory
        Reference for more information.
        """
        command = f"AVPRIN,{key},{effnu}"
        return self.run(command, **kwargs)

    def avres(self, key="", opt="", **kwargs):
        """Specifies how results data will be averaged when PowerGraphics is

        APDL Command: AVRES
        enabled.

        Parameters
        ----------
        key
            Averaging key.

            1 - Average results at all common subgrid locations.

            2 - Average results at all common subgrid locations except where material type
                [MAT] discontinuities exist. This option is the default.

            3 - Average results at all common subgrid locations except where real constant
                [REAL] discontinuities exist.

            4 - Average results at all common subgrid locations except where material type
                [MAT] or real constant [REAL] discontinuities exist.

        opt
            Option to determine how results data are averaged.

            (blank) - Average surface results data using only the exterior element faces (default).

            FULL - Average surface results data using the exterior face and interior element data.

        Notes
        -----
        The AVRES command specifies how results data will be averaged at
        subgrid locations that are common to 2 or more elements.  The command
        is valid only when PowerGraphics is enabled (via the /GRAPHICS,POWER
        command).

        With PowerGraphics active (/GRAPHICS,POWER), the averaging scheme for
        surface data with interior element data included (AVRES,,FULL) and
        multiple facets per edge (/EFACET,2 or /EFACET,4) will yield differing
        minimum and maximum contour values depending on the  Z-Buffering
        options (/TYPE,,6 or /TYPE,,7).  When the Section data is not included
        in the averaging schemes (/TYPE,,7), the resulting absolute value for
        the midside node is significantly smaller.

        PowerGraphics does not average your stresses across discontinuous
        surfaces. The normals for various planes and facets are compared to a
        tolerance to determine continuity. The ANGLE value you specify in the
        /EDGE command is the tolerance for classifying surfaces as continuous
        or "coplanar."

        The command affects nodal solution contour plots (PLNSOL), nodal
        solution printout (PRNSOL), and subgrid solution results accessed
        through the Query Results function (under General Postprocessing) in
        the GUI.

        The command has no effect on the nodal degree of freedom solution
        values (UX, UY, UZ, TEMP, etc.).

        For cyclic symmetry mode-superposition harmonic solutions, AVRES,,FULL
        is not supported. Additionally, averaging does not occur across
        discontinuous surfaces, and the ANGLE value on the /EDGE command has no
        effect.

        The command is also available in /SOLU.
        """
        command = f"AVRES,{key},{opt}"
        return self.run(command, **kwargs)

    def efacet(self, num="", **kwargs):
        """Specifies the number of facets per element edge for PowerGraphics

        APDL Command: /EFACET
        displays.

        .. warning::
           This will not change element plotting within PyMapdl with
           ``eplot(vtk=True)``

        Parameters
        ----------
        num
            Number of facets per element edge for element plots.

            1 - Use 1 facet per edge (default for h-elements).

            2 - Use 2 facets per edge.

            4 - Use 4 facets per edge.

        Notes
        -----
        /EFACET is valid only when PowerGraphics is enabled [/GRAPHICS,POWER],
        except that it can be used in FULL graphics mode for element CONTA174.
        (See the /GRAPHICS command and element CONTA174 in the Element
        Reference for more information.) The /EFACET command is only applicable
        to element type displays.

        /EFACET controls the fineness of the subgrid that is used for element
        plots. The element is subdivided into smaller portions called facets.
        Facets are piecewise linear surface approximations of the actual
        element face. In their most general form, facets are warped planes in
        3-D space.  A greater number of facets will result in a smoother
        representation of the element surface for element plots. /EFACET may
        affect results averaging. See Contour Displays in the Basic Analysis
        Guide for more information.

        For midside node elements, use NUM = 2; if NUM = 1, no midside node
        information is output. For non-midside node elements, NUM should be set
        to 1. See the PLNSOL and PRNSOL commands for more information.

        With PowerGraphics active (/GRAPHICS,POWER), the averaging scheme for
        surface data with interior element data included (AVRES,,FULL) and
        multiple facets per edge (/EFACET,2 or /EFACET,4) will yield differing
        minimum and maximum contour values depending on the  Z-Buffering
        options (/TYPE,,6 or /TYPE,,7).  When the Section data is not included
        in the averaging schemes (/TYPE,,7), the resulting absolute value for
        the midside node is significantly smaller.

        For cyclic symmetry mode-superposition harmonic solutions, only NUM = 1
        is supported in postprocessing.

        Caution:: : If you specify /EFACET,1, PowerGraphics does not plot
        midside nodes. You must use /EFACET,2 to make the nodes visible.

        This command is valid in any processor.
        """
        command = f"/EFACET,{num}"
        return self.run(command, **kwargs)

    def ernorm(self, key="", **kwargs):
        """Controls error estimation calculations.

        APDL Command: ERNORM

        Parameters
        ----------
        key
            Control key:

            ON - Perform error estimation (default). This option is
            not valid for PowerGraphics.

            OFF - Do not perform error estimation.

        Notes
        -----
        Especially for thermal analyses, program speed increases if
        error estimation is suppressed.  Therefore, it might be
        desirable to use error estimation only when needed.  The value
        of the ERNORM key is not saved on file.db. Consequently, you
        need to reissue the ERNORM key after a RESUME if you wish to
        deactivate error estimation again.
        """
        return self.run("ERNORM,%s" % (str(key)), **kwargs)

    def inres(
        self,
        item1="",
        item2="",
        item3="",
        item4="",
        item5="",
        item6="",
        item7="",
        item8="",
        **kwargs,
    ):
        """Identifies the data to be retrieved from the results file.

        APDL Command: INRES

        Parameters
        ----------
        item1, item2, item3, . . . , item8
            Data to be read into the database from the results file.  May
            consist of any of the following labels:

            ALL - All solution items (default).

            BASIC - NSOL, RSOL, NLOAD, STRS, FGRAD, and FFLUX items.

            NSOL - Nodal DOF solution.

            RSOL - Nodal reaction loads.

            ESOL - Element solution items (includes all of the following):

            NLOAD - Element nodal loads.

            STRS - Element nodal stresses.

            EPEL - Element elastic strains.

            EPTH - Element thermal, initial, and swelling strains.

            EPPL - Element plastic strains.

            EPCR - Element creep strains.

            FGRAD - Element nodal gradients.

            FFLUX - Element nodal fluxes.

            MISC - Element miscellaneous data (SMISC and NMISC).

        Notes
        -----
        Identifies the type of data to be retrieved from the results file for
        placement into the database through commands such as SET, SUBSET, and
        APPEND.  INRES is a companion command to the OUTRES command controlling
        data written to the database and the results file.  Since the INRES
        command can only flag data that has already been written to the results
        file, care should be taken when using the OUTRES command to include all
        data you wish to retrieve for postprocessing later on.
        """
        command = (
            f"INRES,{item1},{item2},{item3},{item4},{item5},{item6},{item7},{item8}"
        )
        return self.run(command, **kwargs)

    def layer(self, num="", **kwargs):
        """Specifies the element layer for which data are to be processed.

        APDL Command: LAYER

        Parameters
        ----------
        num
            Layer-processing mode:

            N - The layer number to process. The default value is 0.

            FCMAX - Processes the layer with the largest failure criteria.

        Notes
        -----
        Specifies the element layer for which results data are to be listed,
        plotted, or otherwise processed.

        Applies to stress and strain data for layered elements SHELL163,
        SHELL181, SHELL281, ELBOW290, SOLID185, SOLID186, SOLSH190, SHELL208,
        SHELL209, REINF264, and REINF265; heat flux and heat gradient for
        SHELL131 and SHELL132.

        The SHELL command may then be used (with shell elements) to specify a
        location (TOP, MID, BOT) within the layer for output. (The SHELL
        command does not apply to thermal shell elements SHELL131 and
        SHELL132.) Transverse shear stresses for MID are linearly averaged from
        TOP and BOT, and do not reflect a parabolic distribution. Setting
        KEYOPT(8) = 2 for SHELL181, SHELL281, SHELL208, SHELL209, and ELBOW290
        writes the mid-surface values directly to the results file and yields
        more accurate values than linear averaging.

        Because energy is a per-element quantity, you cannot use this command
        for energy output.

        When using the LAYER command with SHELL181, SOLID185, SOLID186,
        SOLSH190, SHELL208, SHELL209, SHELL281, and ELBOW290, KEYOPT(8) must be
        set to 1 (or 2 for SHELL181, SHELL281, ELBOW290, SHELL208, and
        SHELL209) in order to store results for all layers.

        When NUM = FCMAX, you must provide the failure criterion input. If
        specifying input via the FC command, all structural elements are
        processed. For more information, see the documentation for the FC
        command.

        Use this command with RSYS,LSYS to display results in the layer
        coordinate system for a particular layer.

        For the ANSYS LS-DYNA product, this command works differently than
        described above.  For SHELL163, you must first use EDINT during the
        solution phase to define the integration points for which you want
        output data.  Be aware that the output location for SHELL163 data is
        always at the integration point, so "top" and "bottom" refer to the top
        or bottom integration point, not necessarily the top or bottom surface.
        For more information, see the ANSYS LS-DYNA User's Guide.
        """
        command = f"LAYER,{num}"
        return self.run(command, **kwargs)

    def rsys(self, kcn="", **kwargs):
        """Activates a coordinate system for printout or display of element and

        APDL Command: RSYS
        nodal results.

        Parameters
        ----------
        kcn
            The coordinate system to use for results output:

            0 - Global Cartesian coordinate system (default, except for spectrum analyses).

            1 - Global cylindrical coordinate system.

            2 - Global spherical coordinate system.

            > 10 - Any existing local coordinate system.

            SOLU - Solution coordinate systems. For element quantities, these are the element
                   coordinate system for each element.  For nodal quantities,
                   these are the nodal coordinate systems. If an element or
                   nodal coordinate system is not defined, ANSYS uses the
                   global Cartesian coordinate system. If you issue a LAYER,N
                   command (where N refers to a layer number), the results
                   appear in the layer coordinate system. (SOLU is the default
                   for spectrum analyses.)

            LSYS - Layer coordinate system. For layered shell and solid elements, the results
                   appear in their respective layer coordinate systems.  For a
                   specific layer of interest, issue a LAYER,N command (where N
                   refers to a layer number). If a model has both nonlayered
                   and layered elements, you can use RSYS,SOLU and RSYS,LSYS
                   simultaneously (with RSYS,SOLU applicable to nonlayered
                   elements and RSYS,LSYS applicable to layered elements).  To
                   reverse effects of the LSYS option, issue an RSYS,0 command.
                   LSYS is the default for spectrum analysis.

        Notes
        -----
        The RSYS command activates a coordinate system for printing or
        displaying element results data such as stresses and heat fluxes, and
        nodal results data such as degrees of freedom and reactions. ANSYS
        rotates the results data to the specified coordinate system during
        printout, display, or element table operations (such as PRNSOL, PRESOL,
        PLNSOL, and ETABLE). You can define coordinate systems with various
        ANSYS commands such as LOCAL, CS, CLOCAL, and CSKP.

        If you issue RSYS with KCN > 10 (indicating a local coordinate system),
        and the specified system is subsequently redefined, you must reissue
        RSYS for results to be rotated into the redefined system.

        Note:: : The default coordinate system for certain elements, notably
        shells, is not global Cartesian and is frequently not aligned at
        adjacent elements.

        The use of RSYS,SOLU with these elements can make nodal averaging of
        component element results, such as SX, SY, SZ, SXY, SYZ, and SXZ,
        invalid and is not recommended.

        The RSYS command has no effect on beam or pipe stresses, which ANSYS
        displays (via /ESHAPE,1 and PowerGraphics) in the element coordinate
        system.

        Element results such as stresses and heat fluxes are in the element
        coordinate systems when KCN = SOLU. Nodal requests for element results
        (for example, PRNSOL,S,COMP) average the element values at the common
        node; that is, the orientation of the node is not a factor in the
        output of element quantities.  For nearly all solid elements, the
        default element coordinate systems are parallel to the global Cartesian
        coordinate system. For shell elements and the remaining solid elements,
        the default element coordinate system can differ from element to
        element. For layered shell and layered solid elements, ANSYS initially
        selects the element coordinate system when KCN = SOLU; you can then
        select the layer coordinate system via the LAYER command.

        Nodal results such as degrees of freedom and reactions can be properly
        rotated only if the resulting component set is consistent with the
        degree-of-freedom set at the node. (The degree-of-freedom set at a node
        is determined by the elements attached to the node.) For example, if a
        node does not have a UZ degree of freedom during solution, then any Z
        component resulting from a rotation does not print or display in POST1.
        Therefore, results at nodes with a single degree-of-freedom (UY only,
        for example) should not be rotated; that is, they should be viewed only
        in the nodal coordinate system or a system parallel to the nodal
        system. (The global Cartesian system--the RSYS command default--may not
        be parallel to the nodal system.) Results at nodes with a 2-D degree-
        of-freedom set (UX and UY, for example) should not be rotated out of
        the 2-D plane.

        PowerGraphics

        For PowerGraphics, ANSYS plots PLVECT vector arrow displays (such
        temperature, velocity, and force) in the global Cartesian coordinate
        system (RSYS = 0). Subsequent operations revert to your original
        coordinate system.

        PGR File

        When you generate a .PGR file in SOLUTION, you can use the Results
        Viewer to display your stresses only in the coordinate system in which
        you write your .PGR file. To view stresses in other coordinate systems,
        load your results file into the Results Viewer and regenerate the data.

        Large Deflections

        If large deflection is active (NLGEOM,ON), ANSYS rotates the element
        component result directions by the amount of rigid body rotation.

        ANSYS displays the element component results in the initial global
        coordinate system for the following elements: SHELL181, SHELL281,
        ELBOW290, PLANE182, PLANE183, SOLID185, SOLID186, SOLID187, SOLID272,
        SOLID273, SOLID285, SOLSH190, SHELL208, and SHELL209. All other element
        result transformations are, therefore, also relative to the initial
        global system. Nodal degree-of-freedom results are based on the initial
        (and not the updated) geometry. For all other element types, component
        results displayed in the co-rotated coordinate system include the
        element rigid body rotation from the initial global coordinate system,
        and all other element result transformations are relative to the
        rotated global system.

        LS-DYNA

        You can use the RSYS command to rotate stress data for all explicit
        (ANSYS LS-DYNA) elements except BEAM161, COMBI165, and composite
        SHELL163 (KEYOPT(3) = 1). In models that contain these element types
        combined with other explicit elements, you must unselect the
        unsupported elements before issuing the RSYS command. The command does
        not support strain data for any explicit element types. If you request
        strain results for explicit elements when RSYS is not set to the global
        Cartesian coordinate system (KCN = 0), ANSYS ignores the printing or
        plotting command. (ANSYS always rotates displacements into the results
        coordinate system, independent of the explicit element type.)
        """
        command = f"RSYS,{kcn}"
        return self.run(command, **kwargs)



from . import (
    areas,
    artificially_matched_layers,
    booleans,
    constraint_equations,
    coupled_dof,
    database,
    digitizing,
    element_type,
    elements,
    explicit_dynamics,
    hard_points,
    keypoints,
    lines,
    material_data_tables,
    materials,
    meshing,
    morphing,
    nodes,
    primitives,
    real_constants,
    sections,
    special_purpose,
    status,
    superelements,
    volumes,
)



from . import (
    all_others,
    analysis_options,
    birth_and_death,
    dynamic_options,
    fe_body_loads,
    fe_constraints,
    fe_forces,
    fe_surface_loads,
    gap_conditions,
    inertia,
    load_step_operations,
    load_step_options,
    master_dof,
    miscellaneous_loads,
    multi_field_solver_convergence_controls,
    multi_field_solver_definition_commands,
    multi_field_solver_global_controls,
    multi_field_solver_interface_mapping,
    multi_field_solver_load_transfer,
    multi_field_solver_time_controls,
    nonlinear_options,
    ocean,
    radiosity,
    rezoning,
    solid_body_loads,
    solid_constraints,
    solid_forces,
    solid_surface_loads,
    solution_status,
    spectrum_options,
    twod_to_3d_analysis,
)
